/*
 * p2p iptables match module
 * filipe@rnl.ist.utl.pt
 */


#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/tcp.h>

#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/version.h>


#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
    #undef KERNEL_2_6
#else
    #define KERNEL_2_6
#endif


MODULE_AUTHOR("Filipe Almeida <filipe@rnl.ist.utl.pt>");
MODULE_DESCRIPTION("IP tables p2p match module");
MODULE_LICENSE("GPL");

int
match_http( const unsigned char *data,
             const unsigned char *end);
int
match_edonkey( const unsigned char *data,
             const unsigned char *end);
int
match_dc( const unsigned char *data,
             const unsigned char *end);
int
match_bittorrent( const unsigned char *data,
             const unsigned char *end);
             
static int
match(const struct sk_buff *skb,
      const struct net_device *in,
      const struct net_device *out,
      const void *matchinfo,
      int offset,
#ifndef KERNEL_2_6
      const void *hdr,
      u_int16_t datalen,
#endif
      int *hotdrop)
{
    const struct iphdr *iph = skb->nh.iph;
    const struct tcphdr *tcph;
    const unsigned char *data;
    const unsigned char *end;

#ifdef KERNEL_2_6
    int datalen;
    datalen = skb->len - (iph->ihl<<2);
#endif

    if ( !iph || iph->protocol != IPPROTO_TCP) return 0;

    tcph = (void *)skb->nh.iph + skb->nh.iph->ihl*4;
    data = (const unsigned char *) tcph + tcph->doff * 4;
    end = data + datalen - tcph->doff * 4;

    if (match_http(data, end)) return 1;
    if (match_edonkey(data, end)) return 1;
    if (match_dc(data, end)) return 1;
    if (match_bittorrent(data, end)) return 1;

    return 0;
}

static int
checkentry(const char *tablename,
           const struct ipt_ip *ip,
           void *matchinfo,
           unsigned int matchsize,
           unsigned int hook_mask)
{
    if (matchsize != IPT_ALIGN(0))
        return 0;

    return 1;
}


/*
static struct ipt_match p2p_match
= { { NULL, NULL }, "p2p", &match, &checkentry, NULL, THIS_MODULE };
*/

static struct ipt_match p2p_match = {
    .name        = "p2p",
    .match       = &match,
    .checkentry  = &checkentry,
    .me          = THIS_MODULE,
};

static int __init init(void)
{
    printk(KERN_INFO "Module ipt_p2p loaded.\n");
	return ipt_register_match(&p2p_match);
}

static void __exit fini(void)
{
	ipt_unregister_match(&p2p_match);
}

module_init(init);
module_exit(fini);
